import "Turbine";
import "Turbine.Gameplay";
import "Turbine.UI";
import "Turbine.UI.Lotro";

local step = 100;

ColorBar = class(Turbine.UI.Control);

function ColorBar:Constructor()
	Turbine.UI.Control.Constructor(self);

	self:SetBackColor(Turbine.UI.Color(1, 0.0, 0.0, 0.0));

	self.barColor = { };
	self.backBars = { };
	self.frontBars = { };
end

function ColorBar:SetHeight(h)
	self.h = h;
	Turbine.UI.Control.SetHeight(self, h);
	for i in pairs(self.backBars) do
		self.backBars[i]:SetHeight(h);
		self.frontBars[i]:SetHeight(h);
	end
end

function ColorBar:SetWidth(w)
	self.w = w;
	Turbine.UI.Control.SetWidth(self, w);
	for i in pairs(self.backBars) do
		self.backBars[i]:SetWidth(w);
		self.frontBars[i]:SetWidth(w);
	end
end

function ColorBar:SetSize(w, h)
	self.w = w;
	self.h = h;
	Turbine.UI.Control.SetSize(self, w, h);
	for i in pairs(self.backBars) do
		self.backBars[i]:SetSize(w, h);
		self.frontBars[i]:SetSize(w, h);
	end
end

function ColorBar:SetBarColor(i, c)
	if self.backBars[i] == nil then
		self.backBars[i] = Turbine.UI.Control();
		self.backBars[i]:SetParent(self);
		self.backBars[i]:SetPosition(0, 0);
		self.backBars[i]:SetZOrder(i);
		self.frontBars[i] = Turbine.UI.Control();
		self.frontBars[i]:SetParent(self);
		self.frontBars[i]:SetPosition(0, 0);
		self.frontBars[i]:SetZOrder(step + i);
		if self.h ~= nil then
			self.backBars[i]:SetHeight(self.h);
			self.frontBars[i]:SetHeight(self.h);
		end
	end

	self.barColor[i] = c;

	if c == nil then
		self.backBars[i]:SetVisible(false);
		self.frontBars[i]:SetVisible(false);
	else
		local max = math.max(c.R, c.G, c.B);
		local min = math.min(c.R, c.G, c.B);

		if max == min then
			if max >= 0.5 then
				self.backBars[i]:SetBackColor(Turbine.UI.Color(0.5, c.R - 0.5, c.G - 0.5, c.B - 0.5));
			else
				self.backBars[i]:SetBackColor(Turbine.UI.Color(0.5, c.R + 0.5, c.G + 0.5, c.B + 0.5));
			end
		else
			local v = max;
			local ch = max - min;
			local s = ch / v;

			if v >= 0.5 then
				local dv = 1 / (2 * math.sqrt(1 + ((s * s) / (v * v))));
				local ds = 1 / (2 * math.sqrt(1 + ((v * v) / (s * s))));
				s = s - ds;
				v = v - dv;
			else
				v = 1 - v;
				local dv = 1 / (2 * math.sqrt(1 + ((s * s) / (v * v))));
				local ds = 1 / (2 * math.sqrt(1 + ((v * v) / (s * s))));
				s = s - ds;
				v = 1 - (v - dv);
			end

			local nch = s * v;
			local nmin = v - nch;
			self.backBars[i]:SetBackColor(Turbine.UI.Color(0.5, (((c.R - min) * nch) / ch) + nmin, (((c.G - min) * nch) / ch) + nmin, (((c.B - min) * nch) / ch) + nmin));
		end
		self.frontBars[i]:SetBackColor(c);
	end
end

function ColorBar:SetStats(i, l, c, r, m)
	if self.barColor[i] == nil then
		return;
	end

	if r <= l then
		self.backBars[i]:SetVisible(false);
		self.frontBars[i]:SetVisible(false);
		return;
	end

	m = math.max(m, 1);
	l = math.min(l, m);
	r = math.min(r, m);
	c = math.max(l, c);
	c = math.min(r, c);
	l = math.floor((l * self.w / m) + 0.5)
	c = math.floor((c * self.w / m) + 0.5)
	r = math.floor((r * self.w / m) + 0.5)

	if (r - l) <= 0 then
		self.backBars[i]:SetVisible(false);
		self.frontBars[i]:SetVisible(false);
	else
		self.backBars[i]:SetLeft(l);
		self.backBars[i]:SetWidth(r - l);
		self.backBars[i]:SetVisible(true);
		if (c - l) <= 0 then
			self.frontBars[i]:SetVisible(false);
		else
			self.frontBars[i]:SetLeft(l);
			self.frontBars[i]:SetWidth(c - l);
			self.frontBars[i]:SetVisible(true);
		end
	end
end

function ColorBar:SetMouseVisible(b)
	Turbine.UI.Control.SetMouseVisible(self, b);
	for i in pairs(self.backBars) do
		self.backBars[i]:SetMouseVisible(b);
		self.frontBars[i]:SetMouseVisible(b);
	end
end