ContextMenuItem = class(Turbine.UI.Control);
ContextMenuItem.size = 16;
ContextMenuItem.margin = 2;
ContextMenuItem.wwidth = 8;

function ContextMenuItem:Constructor(parent, window, index)
	Turbine.UI.Control.Constructor(self);

	self.parent = parent;
	self.window = window;
	self.index = index;

	self:SetParent(parent);
	self:SetPosition(0, (index - 1) * (self.size + self.margin * 2));
	self:SetZOrder(1);

	self.icon = Turbine.UI.Control();
	self.icon:SetMouseVisible(false);
	self.icon:SetParent(self);

	self.text = Turbine.UI.Label();
	self.text:SetPosition(self.margin * 3 + self.size, self.margin);
	self.text:SetFont(Turbine.UI.Lotro.Font.Verdana16);
	self.text:SetTextAlignment(Turbine.UI.ContentAlignment.MiddleLeft);
	self.text:SetMouseVisible(false);
	self.text:SetParent(self);
	self.arrow = Turbine.UI.Control();
	self.arrow:SetMouseVisible(false);
	self.arrow:SetParent(self);
	self.arrow:SetBackground("moebius92/PartyFrame/res/expand.tga");
	self.arrow:SetSize(12, 11);
end

function ContextMenuItem:MouseEnter(args)
	self.window:Enter(self.index, self.parent);
end

function ContextMenuItem:MouseLeave(args)
	self.window:Leave(self.index, self.parent);
end

function ContextMenuItem:MouseClick(args)
	self.window:Click(self.index, self.parent);
end

function ContextMenuItem:Set(width, icon, text, submenu)
	if icon == nil then
		self.icon:SetVisible(false);
	else
		self.icon:SetVisible(true);
		self.icon:SetBackground(icon);
		self.icon:SetStretchMode(2);
		self.icon:SetStretchMode(1);
		self.icon:SetSize(self.size, self.size);
		local w, h = self.icon:GetSize();
		self.icon:SetPosition(self.margin * 2 + (self.size - w) / 2, self.margin + (self.size - h) / 2);
	end
	self.text:SetText(text);
	self.text:SetSize(width * self.wwidth, self.size);
	self.text:SetForeColor(Turbine.UI.Color(1.0, 0.75, 0.75, 0.75));
	self.arrow:SetVisible(submenu);
	self.arrow:SetPosition(width * self.wwidth + self.size + self.margin * 4, self.margin + (self.size - self.arrow:GetHeight()) / 2);
	self:SetSize(width * self.wwidth + self.size * 2 + self.margin * 4, self.size + self.margin * 2);
	self:SetVisible(true);
end

function ContextMenuItem:Highlight(h)
	local c = h and 1.0 or 0.75;
	self.text:SetForeColor(Turbine.UI.Color(1.0, c, c, c));
end

ContextMenu = class(Turbine.UI.Window);

function ContextMenu:Constructor(menu, func)
	Turbine.UI.Window.Constructor(self);

	self.menu = menu;
	self.func = func;
	self.index = { };
	self.menus = { };
end

function ContextMenu:Redraw()
	local p = self.menu;
	local i = 0;
	local m, w, h;
	local x = 0;
	local y = 0;
	local mx = 0;
	local my = 0;

	while i <= #self.index do
		if i > 0 then
			x = x + w;
			y = y + h * (self.index[i] - 1);
			p = p[self.index[i]].submenu;
		end

		i = i + 1;
		if self.menus[i] == nil then
			self.menus[i] = Turbine.UI.Lotro.TextBox();
			self.menus[i]:SetReadOnly(true);
			self.menus[i]:SetParent(self);
			self.menus[i].options = { };
			self.menus[i].n = i;
		end
		self.menus[i].menu = p;
		if i > 1 then
			self.menus[i].index = self.index[i - 1];
		end
		self.menus[i]:SetPosition(x, y);

		for _, v in ipairs(self.menus[i].options) do
			v:SetVisible(false);
		end

		m = 0;
		for k, v in ipairs(p) do
			m = math.max(m, string.len(v[1]));
		end

		for k, v in ipairs(p) do
			if self.menus[i].options[k] == nil then
				self.menus[i].options[k] = ContextMenuItem(self.menus[i], self, k);
			end
			self.menus[i].options[k]:Set(m, v[2], v[1], v.submenu ~= nil);
		end

		w, h = self.menus[i].options[1]:GetSize();
		self.menus[i]:SetSize(w, h * #p);
		mx = math.max(x + w, mx);
		my = math.max(y + h * #p, my);
	end

	self:SetSize(mx, my);
end

function ContextMenu:Enter(i, menu)
	local index = { };

	for i = 2, menu.n, 1 do
		table.insert(index, self.menus[i].index);
	end
	if menu.menu[i].submenu ~= nil then
		table.insert(index, i);
	end

	if #index == #self.index then
		for k, v in ipairs(self.index) do
			if v ~= index[k] then
				self.index = index;
				self:Redraw();
				break;
			end
		end
	else
		self.index = index;
		self:Redraw();
	end
	menu.options[i]:Highlight(true);
end

function ContextMenu:Leave(i, menu)
	menu.options[i]:Highlight(false);
end

function ContextMenu:Click(i, menu)
	local index = { };

	if menu.menu[i].submenu ~= nil then
		for i = 2, menu.n, 1 do
			table.insert(index, self.menus[i].index);
		end
		table.insert(index, i);

		if #index == #self.index then
			table.remove(index, #index);
		end
		self.index = index;
		self:Redraw();
	else
		local f = self.func;

		if menu.menu[i].func ~= nil then
			f = menu.menu[i].func;
		end
		if f(menu.menu[i]) then
			self:SetVisible(false);
		end
	end
	menu.options[i]:Highlight(true);
end

function ContextMenu:SetPosition(x, y)
	Turbine.UI.Window.SetPosition(self, x, y);
	self:SetVisible(true);
	self:Redraw();
	self:Activate();
end

function ContextMenu:Deactivated(args)
	self:SetVisible(false);
end