PartyFrameCommand = class(Turbine.ShellCommand);

function PartyFrameCommand:Constructor(plugin)
	Turbine.ShellCommand.Constructor(self);

	self.plugin = plugin;
	self.window = plugin.window;
end

function PartyFrameCommand:Execute(command, arguments)
	local args = snarfargs(arguments);

	if args[1] == S.CommandClear then
		self.window:ClearRolls();
		for _, player in pairs(self.window.players) do
			player.rolls = { };
			player.rollOrder = nil;
			player:UpdateName();
		end
	elseif args[1] == S.CommandLoad then
		if args[2] == nil then
			local setups = self.window.options:ListSavedControls();

			if #setups == 0 then
				Turbine.Shell.WriteLine(S.CommandNotFound);
			else
				table.sort(setups);
				for _, setup in ipairs(setups) do
					Turbine.Shell.WriteLine(setup);
				end
			end
		elseif self.window.options:LoadSavedControls(args[2]) then
			Turbine.Shell.WriteLine(self.plugin:GetName() .. ": " .. args[2] .. " " .. S.CommandLoaded .. ".");
		else
			Turbine.Shell.WriteLine(self.plugin:GetName() .. ": " .. args[2] .. " " .. S.CommandNotFound .. ".");
		end
	elseif args[1] == S.CommandPoll then
		if tonumber(args[2]) ~= nil then
			self.window:SetEffectPoll(tonumber(args[2]));
		end
		if self.window.effectPoll == 0 then
			Turbine.Shell.WriteLine(self.plugin:GetName() .. ": " .. S.CommandPollOff .. ".");
		else
			Turbine.Shell.WriteLine(self.plugin:GetName() .. ": " .. self.window.effectPoll .. " " .. S.CommandPollInterval .. ".");
		end
	elseif args[1] == S.CommandOptions then
		Turbine.PluginManager.ShowOptions(self.plugin);
	else
		Turbine.Shell.WriteLine(S.CommandHelpL);
	end
end

function PartyFrameCommand:GetShortHelp()
	return self.plugin:GetName();
end