import "moebius92.PartyFrame.ContextMenu";

PartyFrameOptionsEffectText = class(Turbine.UI.Lotro.TextBox);

function PartyFrameOptionsEffectText:Constructor(parent)
	Turbine.UI.Lotro.TextBox.Constructor(self);

	self.parent = parent;
	self:SetParent(parent);
	self:SetMultiline(false);
	self:SetTextAlignment(Turbine.UI.ContentAlignment.MiddleLeft);
	self:SetFont(Turbine.UI.Lotro.Font.Verdana16);
	self:SetZOrder(0);
	self:SetMouseVisible(true);
end

function PartyFrameOptionsEffectText:FocusGained(args)
	self.value = self:GetText();
end

function PartyFrameOptionsEffectText:FocusLost(args)
	self:SetText(self.value);
	self.value = nil;
end

function PartyFrameOptionsEffectText:KeyDown(args)
	if args.Action == 162 then
		local name = self:GetText();
		local option = self.parent.parent;

		if self.parent.group == nil then
			table.insert(option.controls[option.index].color, { color = { }, [1] = { name = name, debuff = false } });
		else
			option.controls[option.index].color[self.parent.group][self.parent.index].name = name;
			if not option.controls[option.index].missing then
				table.sort(option.controls[option.index].color[self.parent.group], function(a, b) return a.name < b.name; end);
			end
		end
		option:Changed();
		self.value = name;

		self.parent:Focus();
	end
end

PartyFrameOptionsEffect = class(Turbine.UI.Control)

function PartyFrameOptionsEffect:Constructor(parent, icon, debuff, name, ro)
	Turbine.UI.Control.Constructor(self);

	self.parent = parent;
	self.dragging = false;
	self:SetParent(parent);
	self:SetMouseVisible(true);

	self.icon = Turbine.UI.Control();
	self.icon:SetParent(self);
	self.icon:SetMouseVisible(false);
	self.icon:SetBackColor(Turbine.UI.Color(1.0, 0.0, 0.0, 0.0));
	self.icon:SetBlendMode(Turbine.UI.BlendMode.Overlay);
	self.icon:SetPosition(1, 1);
	self.textbox = PartyFrameOptionsEffectText(self);
	self.dd = Turbine.UI.Button();
	self.dd:SetBackground("moebius92/PartyFrame/res/dropdown.tga");
	self.dd:SetBlendMode(Turbine.UI.BlendMode.Overlay);
	self.dd:SetZOrder(1);
	self.dd:SetSize(self.parent.size - 4, self.parent.size);
	self.dd:SetParent(self);
	self.dd.parent = self;
	self.dd.MouseClick = function(self, args)
		PartyFrameOptionsEffect.textTarget = self.parent;
		self.parent.textMenu:SetPosition(self.parent.textbox:PointToScreen(0, self.parent.textbox:GetHeight()));
	end;
	self:SetIcon(icon);
	self:SetDebuff(debuff);
	self:SetName(name, ro);
end

function PartyFrameOptionsEffect:MouseDown(args)
	if args.Button == Turbine.UI.MouseButton.Left then
		local x, y = Turbine.UI.Display.GetMousePosition();

		self.dragWindow:SetBackground(self.image);
		if self.image ~= nil then
			self.dragWindow:SetBackColorBlendMode(Turbine.UI.BlendMode.Screen);
		else
			self.dragWindow:SetBackColorBlendMode(Turbine.UI.BlendMode.Normal);
		end
		self.dragWindow:Activate();
		self.dragWindow:SetZOrder(5000);
		self.dragWindow:SetVisible(true);
		self.dragWindow:SetPosition(x - 16, y - 16);
		self.dragging = true;
	end
end

function PartyFrameOptionsEffect:MouseMove(args)
	if self.dragging then
		local x, y = Turbine.UI.Display.GetMousePosition();
		self.dragWindow:SetPosition(x - 16, y - 16);
	end
end

function PartyFrameOptionsEffect:MouseEnter(args)
	if self.parent.dragSource ~= nil and self.parent.dragTime == Turbine.Engine.GetGameTime() and self.parent.index ~= 0 then
		local tg = self.group;
		local ti = self.index;
		local sg = self.parent.dragSource.group;
		local si = self.parent.dragSource.index;
		local x, y = self:GetMousePosition();
		local d = (y > ((self.parent.size + 2) / 2)) and 1 or 0;

		if self == self.parent.dragSource then
			return;
		end

		if sg == nil then
			if tg == nil then
				table.insert(self.parent.controls[self.parent.index].color, { color = { }, [1] = { debuff = false, name = "" } });
			else
				table.insert(self.parent.controls[self.parent.index].color[tg], ti + d, { debuff = false, name = "" });
			end
		elseif tg == nil then
			table.insert(self.parent.controls[self.parent.index].color, { color = { }, [1] = self.parent.controls[self.parent.index].color[sg][si] });
			self.parent:RemoveEffect(sg, si);
		else
			if sg ~= tg or si < ti then
				table.insert(self.parent.controls[self.parent.index].color[tg], ti + d, self.parent.controls[self.parent.index].color[sg][si]);
				self.parent:RemoveEffect(sg, si);
			else
				table.insert(self.parent.controls[self.parent.index].color[tg], ti + d, self.parent:RemoveEffect(sg, si));
			end
			if not self.parent.controls[self.parent.index].missing then
				table.sort(self.parent.controls[self.parent.index].color[tg], function(a, b) return a.name < b.name; end);
			end
		end
		self.parent:Changed();
	end
end

function PartyFrameOptionsEffect:MouseUp(args)
	if args.Button == Turbine.UI.MouseButton.Left then
		self.dragWindow:SetVisible(false);
		self.dragging = false;
		self.parent.dragSource = self;
		self.parent.dragTime = Turbine.Engine.GetGameTime();
	end
end

function PartyFrameOptionsEffect:MouseClick(args)
	if args.Button == Turbine.UI.MouseButton.Right then
		PartyFrameOptionsEffect.iconTarget = self;
		self.options.Debuff:SetChecked(self.debuff);
		if self.textbox:IsReadOnly() then
			self.options.Cut:SetEnabled(true);
			self.options.Copy:SetEnabled(true);
			self.options.Paste:SetEnabled(false);
			self.options.Insert:SetEnabled(false);
			self.options.Delete:SetEnabled(true);
			self.options.Debuff:SetEnabled(false);
			self.options.Save:SetEnabled(true);
			self.options.NewGroup:SetEnabled(false);
		else
			self.options.Cut:SetEnabled(self.group ~= nil);
			self.options.Copy:SetEnabled(self.group ~= nil);
			self.options.Paste:SetEnabled(self.clipboard ~= nil);
			self.options.Insert:SetEnabled(true);
			self.options.Delete:SetEnabled(self.group ~= nil);
			self.options.Debuff:SetEnabled(self.group ~= nil);
			self.options.Save:SetEnabled(self.group ~= nil);
			self.options.NewGroup:SetEnabled(self.group ~= nil and self.index > 1);
		end
		self.iconMenu:ShowMenu();
	end
end

function PartyFrameOptionsEffect:SetIcon(icon)
	local w, h = self:GetSize();

	self.icon:SetBackground(icon);
	self.image = icon;
	if icon == nil then
		self.icon:SetBackColorBlendMode(Turbine.UI.BlendMode.Normal);
	else
		self.icon:SetBackColorBlendMode(Turbine.UI.BlendMode.Screen);
	end
end

function PartyFrameOptionsEffect:SetName(name, ro)
	self.textbox:SetText(name);
	self.textbox:SetReadOnly(ro);
	self:SetDebuff(self.debuff);
	self.dd:SetVisible(not ro);
end

function PartyFrameOptionsEffect:SetDebuff(debuff)
	local name = self.textbox:GetText();

	self.debuff = debuff;
	if name ~= nil and string.sub(name, 1, 1) == "[" then
		self.textbox:SetForeColor(Turbine.UI.Color(1.0, 1.0, 1.0, 1.0));
	else
		if debuff then
			self.textbox:SetForeColor(Turbine.UI.Color(1.0, 1.0, 0.0, 0.0));
		else
			self.textbox:SetForeColor(Turbine.UI.Color(1.0, 0.0, 1.0, 0.0));
		end
	end
end

function PartyFrameOptionsEffect:SizeChanged(args)
	local w, h = self:GetSize();

	w = math.max(20, w - (h + 2));
	self.icon:SetSize(h - 2, h - 2);
	self.textbox:SetSize(w, h - 2);
	self.textbox:SetPosition(h + 1, 1);
	self.dd:SetPosition(w + 4, 1);
end

function PartyFrameOptionsEffect:SetIndex(group, index)
	self.group = group;
	self.index = index;
end

PartyFrameOptionsEffect.dragWindow = Turbine.UI.Window();
PartyFrameOptionsEffect.dragWindow:SetSize(32, 32);
PartyFrameOptionsEffect.dragWindow:SetBackColor(Turbine.UI.Color(1.0, 0.0, 0.0, 0.0));
PartyFrameOptionsEffect.dragWindow:SetOpacity(0.7);

PartyFrameOptionsEffect.clipboard = nil;
PartyFrameOptionsEffect.options = { };
PartyFrameOptionsEffect.iconMenu = Turbine.UI.ContextMenu();
local items = PartyFrameOptionsEffect.iconMenu:GetItems();
PartyFrameOptionsEffect.options.Cut = Turbine.UI.MenuItem(S.EffectCMCut);
PartyFrameOptionsEffect.options.Cut.Click = function(self, args)
	local target = PartyFrameOptionsEffect.iconTarget;
	PartyFrameOptionsEffect.clipboard = target.parent:RemoveEffect(target.group, target.index);
	target.parent:Changed();
end;
items:Add(PartyFrameOptionsEffect.options.Cut);
PartyFrameOptionsEffect.options.Copy = Turbine.UI.MenuItem(S.EffectCMCopy);
PartyFrameOptionsEffect.options.Copy.Click = function(self, args)
	local target = PartyFrameOptionsEffect.iconTarget;
	if target.parent.index == 0 then
		PartyFrameOptionsEffect.clipboard = tabledeepcopy(target.parent.controls[target.parent.index][target.index]);
	else
		PartyFrameOptionsEffect.clipboard = tabledeepcopy(target.parent.controls[target.parent.index].color[target.group][target.index]);
	end
end;
items:Add(PartyFrameOptionsEffect.options.Copy);
PartyFrameOptionsEffect.options.Paste = Turbine.UI.MenuItem(S.EffectCMPaste);
PartyFrameOptionsEffect.options.Paste.Click = function(self, args)
	local target = PartyFrameOptionsEffect.iconTarget;
	if target.group == nil then
		table.insert(target.parent.controls[target.parent.index].color, { color = { }, [1] = tabledeepcopy(PartyFrameOptionsEffect.clipboard) });
	else
		table.insert(target.parent.controls[target.parent.index].color[target.group], target.index, tabledeepcopy(PartyFrameOptionsEffect.clipboard));
	end
	target.parent:Changed();
end;
items:Add(PartyFrameOptionsEffect.options.Paste);
PartyFrameOptionsEffect.options.Delete = Turbine.UI.MenuItem(S.EffectCMDelete);
PartyFrameOptionsEffect.options.Delete.Click = function(self, args)
	local target = PartyFrameOptionsEffect.iconTarget;
	target.parent:RemoveEffect(target.group, target.index);
	target.parent:Changed();
end;
items:Add(PartyFrameOptionsEffect.options.Delete);
PartyFrameOptionsEffect.options.Insert = Turbine.UI.MenuItem(S.EffectCMInsert);
PartyFrameOptionsEffect.options.Insert.Click = function(self, args)
	local target = PartyFrameOptionsEffect.iconTarget;
	if target.group == nil then
		table.insert(target.parent.controls[target.parent.index].color, { color = { }, [1] = { debuff = false, name = "" } });
	else
		table.insert(target.parent.controls[target.parent.index].color[target.group], target.index, { debuff = false, name = "" });
	end
	target.parent:Changed();
end;
items:Add(PartyFrameOptionsEffect.options.Insert);
PartyFrameOptionsEffect.options.Debuff = Turbine.UI.MenuItem(S.EffectCMDebuff);
PartyFrameOptionsEffect.options.Debuff.Click = function(self, args)
	local target = PartyFrameOptionsEffect.iconTarget;
	target.parent.controls[target.parent.index].color[target.group][target.index].debuff = not target.parent.controls[target.parent.index].color[target.group][target.index].debuff;
	target.parent:Changed();
end;
items:Add(PartyFrameOptionsEffect.options.Debuff);
PartyFrameOptionsEffect.options.Save = Turbine.UI.MenuItem(S.EffectCMSave);
PartyFrameOptionsEffect.options.Save.Click = function(self, args)
	local target = PartyFrameOptionsEffect.iconTarget;
	local targetEffect;

	if target.parent.index == 0 then
		targetEffect = target.parent.controls[target.parent.index][target.index];
	else
		targetEffect = target.parent.controls[target.parent.index].color[target.group][target.index];
	end
	for _, effect in pairs(target.parent.logSaved) do
		if effect.name == targetEffect.name and effect.debuff == targetEffect.debuff and effect.icon == targetEffect.icon then
			return;
		end
	end
	table.insert(target.parent.logSaved, tabledeepcopy(targetEffect));
	target.parent:SaveLogSaved();
end;
items:Add(PartyFrameOptionsEffect.options.Save);
PartyFrameOptionsEffect.options.NewGroup = Turbine.UI.MenuItem(S.EffectCMNewGroup);
PartyFrameOptionsEffect.options.NewGroup.Click = function(self, args)
	local target = PartyFrameOptionsEffect.iconTarget;
	local n = #target.parent.controls[target.parent.index].color[target.group];

	table.insert(target.parent.controls[target.parent.index].color, target.group + 1, { color = { } });
	for i = target.index, n, 1 do
		table.insert(target.parent.controls[target.parent.index].color[target.group + 1], target.parent.controls[target.parent.index].color[target.group][i]);
		target.parent.controls[target.parent.index].color[target.group][i] = nil;
	end
	target.parent:Changed();
end;
items:Add(PartyFrameOptionsEffect.options.NewGroup);

PartyFrameOptionsEffect.textMenu = ContextMenu(
	{
		{ S.EffectFieldDebuff, submenu = {
			{ S.SE_Disease }, { S.SE_Fear }, { S.SE_Poison }, { S.SE_Wound }, { S.SE_Cure }, { S.SE_ICure },
			{ S.SE_DiseaseCure }, { S.SE_FearCure }, { S.SE_PoisonCure }, { S.SE_WoundCure },
			{ S.SE_DiseaseICure }, { S.SE_FearICure }, { S.SE_PoisonICure }, { S.SE_WoundICure } }
		},
		{ S.EffectFieldClass, submenu = {
			{ S.SE_Beorning, 0x4115370D },
			{ S.SE_Burglar, 0x41007DEB }, { S.SE_Captain, 0x41007DE7 }, { S.SE_Champion, 0x41007DEC },
			{ S.SE_Guardian, 0x41007DE8 }, { S.SE_Hunter, 0x41007DEA }, { S.SE_Minstrel, 0x41007DE6 }, 
			{ S.SE_LoreMaster, 0x41007DE9 }, { S.SE_RuneKeeper, 0x410EA238 }, { S.SE_Warden, 0x410EA239 } }
		},
		{ S.EffectFieldStatus, submenu = {
			{ S.SE_LinkDead_true }, { S.SE_LinkDead_false },
			{ S.SE_Voice_true }, { S.SE_Voice_false },
			{ S.SE_Talking_true }, { S.SE_Talking_false },
			{ S.SE_Leader_true }, { S.SE_Leader_false },
			{ S.SE_Target_true }, { S.SE_Target_false },
			{ S.SE_Morale }, { S.SE_Power }, { S.SE_Default }, { S.SE_Player }, { S.SE_Pet } }
		}
	},
	function(item)
		local target = PartyFrameOptionsEffect.textTarget;
		local option = target.parent;

		target.textbox:SetText(item[1]);
		if target.group == nil then
			table.insert(option.controls[option.index].color, { color = { }, [1] = { name = item[1], debuff = true } });
		else
			option.controls[option.index].color[target.group][target.index].name = item[1];
			option.controls[option.index].color[target.group][target.index].debuff = true;
			if not option.controls[option.index].missing then
				table.sort(option.controls[option.index].color[target.group], function(a, b) return a.name < b.name; end);
			end
		end
		option:Changed();
		return true;
	end
);