import "moebius92.PartyFrame.PartyFrameFellowshipTab";
import "moebius92.PartyFrame.PartyFrameOptions";
import "moebius92.PartyFrame.PartyMemberWindow";

local fakeParty = false;

function IsSafeY(self, Y)
	if self.safeY ~= nil then
		if self.safeY > 0 then
			if Y > 0 and Y < self.safeY then
				self.safeY = Y;
				return false;
			end
		elseif self.safeY < 0 then
			if Y < 0 and Y > self.safeY then
				self.safeY = Y;
				return false;
			end
		end
		self.safeY = nil;
	end
	return true;
end

PartyFrameWindow = class(Turbine.UI.Window);

function PartyFrameWindow:Constructor(version)
	Turbine.UI.Window.Constructor(self);

	self.size = 22;
	self.hide = false;
	self.move = false;
	self.dragging = false;
	self.reload = false;
	self.restart = false;

	self.maxMorale = 1;
	self.maxPower = 1;
	self.effectPollDisable = 0;
	self.player = Turbine.Gameplay.LocalPlayer.GetInstance();
	self.class = self.player:GetClass();
	self.effects = self.player:GetEffects();
	self.target = self.player:GetTarget();

	self.w = Turbine.UI.Display:GetWidth();
	self.h = Turbine.UI.Display:GetHeight();
	self.version = version;

	self:LoadSettings();

	self.players = { };
	self.actors = { };
	self.fellowtabs = { };
	self.fellowship = { };
	self.grid = { };

	self.effectControl = { };
	self.moraleControl = { };
	self.powerControl = { };
	self.options = PartyFrameOptions(self);

	self:SetWantsKeyEvents(true);
	self:SetMouseVisible(true);

	self:SetBackColor(Turbine.UI.Color(1.0, 1.0, 1.0, 1.0));
	self.l = Turbine.UI.Control();
	self.l:SetParent(self);
	self.l:SetSize(self.size, self.size);
	self.l:SetBackColor(Turbine.UI.Color(1.0, 1.0, 1.0, 1.0));
	self.l:SetBackColorBlendMode(Turbine.UI.BlendMode.Multiply);
	self.l:SetMouseVisible(false);
	self.r = Turbine.UI.Control();
	self.r:SetParent(self);
	self.r:SetSize(self.size, self.size);
	self.r:SetBackColor(Turbine.UI.Color(1.0, 1.0, 1.0, 1.0));
	self.r:SetBackColorBlendMode(Turbine.UI.BlendMode.Multiply);
	self.r:SetMouseVisible(false);
	self:Orient();

	self.memberOptions = { };
	self.memberMenu = Turbine.UI.ContextMenu();
	local items = self.memberMenu:GetItems();
	local f = function(self, args)
		self.parent.memberSelect.showPet = self.index;
		self.parent:PartyMemberChanged();
	end;
	for i, option in ipairs({ S.MemberVitalCMHidePet, S.MemberVitalCMPetBelow, S.MemberVitalCMPetAbove, S.MemberVitalCMPetLeft, S.MemberVitalCMPetRight }) do
		self.memberOptions[i] = Turbine.UI.MenuItem(option);
		self.memberOptions[i].parent = self;
		self.memberOptions[i].index = i;
		self.memberOptions[i].Click = f;
		items:Add(self.memberOptions[i]);
	end

	self.tabOptions = { };
	self.tabMenu = Turbine.UI.ContextMenu();
	local items = self.tabMenu:GetItems();
	local f = function(self, args)
		self.parent:Orient(self.arrange, self.orient);
	end;
	self.tabOptions.options = { };

	local arrangeOptions = { 1, 1, 2, 2, 2, 2 };
	local orientOptions = { 1, 2, 1, 2, 3, 4 };

	for i, option in ipairs({ S.PartyFrameCMLineUp, S.PartyFrameCMLineDown, S.PartyFrameCMGridUp, S.PartyFrameCMGridDown, S.PartyFrameCMGridRight, S.PartyFrameCMGridLeft }) do
		self.tabOptions.options[i] = Turbine.UI.MenuItem(option);
		self.tabOptions.options[i].parent = self;
		self.tabOptions.options[i].arrange = arrangeOptions[i];
		self.tabOptions.options[i].orient = orientOptions[i];
		self.tabOptions.options[i].Click = f;
		items:Add(self.tabOptions.options[i]);
	end

	self.partyHandle = { };
	self:UpdateParty();

	self.effectsHandle = { };
	self.playerHandle = { };
	self.playerHandle["PartyChanged"] = moebius92.Handlers.AddEventHandler(self.player, "PartyChanged", function(sender, args) self:UpdateParty(); end);

	self.chatHandle = moebius92.Handlers.AddChatHandler(function(m) self:ChatMessage(m) end, Turbine.ChatType.Standard, nil);

	self.effectsHandle["EffectAdded"] = moebius92.Handlers.AddEventHandler(self.effects, "EffectAdded", function(sender, args) self:LogEffect(self.effects:Get(args.Index)); end);
	self:PetChanged();
	self.playerHandle["PetChanged"] = moebius92.Handlers.AddEventHandler(self.player, "PetChanged", function(sender, args) self:PetChanged(); end);

	self.inCombat = self.player:IsInCombat();
	self.playerHandle["InCombatChanged"] = moebius92.Handlers.AddEventHandler(self.player, "InCombatChanged", function(sender, args) self:CombatChange(); end);

	self.playerHandle["TargetChanged"] = moebius92.Handlers.AddEventHandler(self.player, "TargetChanged",
function(sender, args)
	self.target = self.player:GetTarget();
	for _, actor in pairs(self.actors) do
		actor:CheckTarget();
	end
end
	                                   );
end

function PartyFrameWindow:LoadSettings()
	local settings = Turbine.PluginData.Load(Turbine.DataScope.Character, "PartyFrame");
	if type(settings) == "string" then
		settings = loadstring("return " .. settings .. ";")();
	end

	self.layout = layouts[1];
	self.x = self.w - self.layout.width;
	self.y = self.h / 2 - self.size;
	self.orient = 1;
	self.arrange = 1;
	self.effectPoll = 1.0;

	if settings ~= nil then
		local v = snarfversion(settings.version);

		if versioncmp(settings.version, "0.6") >= 0 then
			for _, layout in ipairs(layouts) do
				if layout.name == settings.layout then
					self.layout = layout;
					break;
				end
			end

			if self.layout == nil then
				self.layout = layouts[1];
			end

--			self.x = settings.x;
--			self.y = settings.y;
			self.orient = settings.orient;
			self.arrange = settings.arrange;
		end

		if versioncmp(settings.version, "0.8") >= 0 then
			self.effectPoll = settings.effectPoll;
		end
	end

	Turbine.UI.Lotro.LotroUI.SetEnabled(Turbine.UI.Lotro.LotroUIElement.Party, self.layout.overlay);
end

function PartyFrameWindow:SaveSettings()
	local settings = { };

	settings.x = self.x;
	settings.y = self.y;
	settings.orient = self.orient;
	settings.arrange = self.arrange;
	settings.layout = self.layout.name;
	settings.effectPoll = self.effectPoll;
	settings.version = self.version;
	Turbine.PluginData.Save(Turbine.DataScope.Character, "PartyFrame", serialize(settings));
end

function PartyFrameWindow:Update()
	if self.restart then
		Turbine.PluginManager.UnloadScriptState("PartyFrameReloader");
		self.restart = false;
	end

	if self.nextUpdate <= Turbine.Engine.GetGameTime() and not self.options:IsDisplayed() then
		Turbine.PluginManager.LoadPlugin("PartyEffects");
		for _, command in pairs(Turbine.Shell.GetCommands()) do
			if string.sub(command, 1, 8) == "effects-" then
				command = string.sub(command, 9);
				local t = { };
				for id in string.gmatch(command, "%d+") do
					t[tonumber(id)] = true;
				end
				for _, actor in pairs(self.actors) do
					actor:CheckEffects(t);
				end
				break;
			end
		end
		Turbine.PluginManager.UnloadScriptState("PartyEffects");
		self.nextUpdate = self.nextUpdate + self.effectPoll;
	end
end

function PartyFrameWindow:SetEffectPoll(t)
	self.effectPoll = t;
	self.nextUpdate = Turbine.Engine.GetGameTime();
	self:SetWantsUpdates(#self.actors > 0 and self.effectPoll > 0);
	self:SaveSettings();
end

function PartyFrameWindow:KeyDown(args)
	if args.Action == 0x1000007B then
		self:SetMove(not self.move);
	elseif args.Action == 0x100000B3 then
		self:SetHide(not self.hide);
	end
end

function PartyFrameWindow:SetMove(m)
	self.move = m;
	if self.move then
		for _, player in pairs(self.players) do
			player.select:SetMouseVisible(false);
		end
	else
		if self.dragging then
			self.dragging = false;
			self:SaveSettings();
		end
		for _, fellowtab in pairs(self.fellowtabs) do
			fellowtab.dragging = false;
		end
		for _, player in pairs(self.players) do
			player.dragging = false;
			player.select:SetMouseVisible(true);
		end
	end
	self:Arrange();
	self:SetVisible(not self.hide and self.move);
end

function PartyFrameWindow:SetHide(h)
	self.hide = h;
	for _, actor in pairs(self.actors) do
		actor:SetVisible(not self.hide);
	end
	self:SetVisible(not self.hide and self.move);
end

function PartyFrameWindow:MouseDown(args)
	if args.Button == Turbine.UI.MouseButton.Left then
		self.dragStartX = args.X;
		self.dragStartY = args.Y;
		self.dragging = true;
	end
end

function PartyFrameWindow:MouseMove(args)
	if self.dragging then
		self.x = self.x + (args.X - self.dragStartX);
		self.y = self.y + (args.Y - self.dragStartY);

		self.x = math.max(0, math.min(self.w - self.layout.width, self.x));
		self:Arrange();
	end
end

function PartyFrameWindow:MouseUp(args)
	if args.Button == Turbine.UI.MouseButton.Left and self.dragging then
		self.dragging = false;
		self:MaxOffset();
		self:SaveSettings();
	end
end

function PartyFrameWindow:MouseClick(args)
	if args.Button == Turbine.UI.MouseButton.Right then
		for i = 1, #self.tabOptions.options, 1 do
			self.tabOptions.options[i]:SetChecked(self.arrange == self.tabOptions.options[i].arrange and self.orient == self.tabOptions.options[i].orient);
		end
		self.tabMenu:ShowMenu();
	end
end

function PartyFrameWindow:MaxOffset()
	if self.majorX < 0 then
		self.maxOffset = math.max(0, math.floor(self.x / -self.majorX));
	elseif self.majorX > 0 then
		self.maxOffset = math.max(0, math.floor((self.w - self.x) / self.majorX));
	elseif self.majorY < 0 then
		self.maxOffset = math.max(0, math.floor(self.y / -self.majorY));
	elseif self.majorY > 0 then
		self.maxOffset = math.max(0, math.floor((self.h - self.y) / self.majorY));
	end
end

function PartyFrameWindow:Orient(arrange, orient)
	if arrange ~= nil and self.arrange ~= arrange then
		self.arrange = arrange;
		self.orient = orient;
		self.fellowship = { };
		self.grid = { };

		if arrange == 1 then
			for _, player in pairs(self.players) do
				if player.showPet ~= 1 then
					player.showPet = 2;
				end
			end
		else
			local x = 1;
			if self.orient <= 2 then
				if self.x < (self.w / 2) then
					x = 5;
				else
					x = 4;
				end
			else
				if self.y < (self.h / 2) then
					x = 2;
				else
					x = 3;
				end
			end
			for _, player in pairs(self.players) do
				if player.showPet ~= 1 then
					player.showPet = x;
				end
			end
		end
	elseif orient ~= nil and self.orient ~= orient then
		self.orient = orient;
	end

	self.dX = 0;
	self.dY = 0;
	self.majorX = 0;
	self.majorY = 0;
	self.minorX = 0;
	self.minorY = 0;

	if self.arrange == 1 then
		if self.orient == 1 then
			self.dY = self.layout.height;
			self.majorY = -self.layout.height;
			self.minorX = self.layout.width;
		else
			self.dY = -self.size;
			self.majorY = self.layout.height;
			self.minorX = self.layout.width;
		end
	else
		if self.orient == 1 then
			self.dY = self.layout.height;
			self.majorY = -self.layout.height;
			self.minorX = self.layout.width;
		elseif self.orient == 2 then
			self.dY = -self.size;
			self.majorY = self.layout.height;
			self.minorX = self.layout.width;
		elseif self.orient == 3 then
			self.dX = -self.size;
			self.majorX = self.layout.width;
			self.minorY = self.layout.height;
		else
			self.dX = self.layout.width;
			self.majorX = -self.layout.width;
			self.minorY = self.layout.height;
		end
	end

	if self.orient == 1 then
		self.l:SetBackground("moebius92/PartyFrame/res/ll.tga");
		self.r:SetBackground("moebius92/PartyFrame/res/lr.tga");
	elseif self.orient == 2 then
		self.l:SetBackground("moebius92/PartyFrame/res/ul.tga");
		self.r:SetBackground("moebius92/PartyFrame/res/ur.tga");
	elseif self.orient == 3 then
		self.l:SetBackground("moebius92/PartyFrame/res/ul.tga");
		self.r:SetBackground("moebius92/PartyFrame/res/ll.tga");
	else
		self.l:SetBackground("moebius92/PartyFrame/res/ur.tga");
		self.r:SetBackground("moebius92/PartyFrame/res/lr.tga");
	end
	local w = math.max(self.size, self.minorX);
	local h = math.max(self.size, self.minorY);
	self:SetSize(w, h);
	self.l:SetPosition(0, 0);
	self.r:SetPosition(math.max(0, self.minorX - self.size), math.max(0, self.minorY - self.size));
	if (self.x + self.dX) >= self.w then
		self.x = (self.w - 1) - self.dX;
	elseif (self.x + self.dX + w) <= 0 then
		self.x = 1 - (self.dX + w);
	end
	if (self.y + self.dY) >= self.h then
		self.y = (self.h - 1) - self.dY;
	elseif (self.y + self.dY + h) <= 0 then
		self.y = 1 - (self.dY + h);
	end

	self:MaxOffset();
	self:SaveSettings();
	self:PartyMemberChanged();
end

function PartyFrameWindow:MoveMember(id, dx, dy)
	local x = self.players[id].indexX;
	local y = self.players[id].indexY;

	if self.arrange == 1 then
		if dx == 0 then
			if dy < 0 then
				if y > 1 then
					self.fellowship[x][y] = self.fellowship[x][y - 1];
					self.fellowship[x][y - 1] = id;
				else
					table.remove(self.fellowship[x], y);
					x = x - orientDFellowship[self.orient];
					if self.fellowship[x] ~= nil then
						table.insert(self.fellowship[x], id);
					elseif x == 0 then
						table.insert(self.fellowship, 1, { [1] = id });
					else
						self.fellowship[x] = { [1] = id };
					end
				end
				self:GridArrange();
				return true;
			elseif dy > 0 then
				if y < #self.fellowship[x] then
					self.fellowship[x][y] = self.fellowship[x][y + 1];
					self.fellowship[x][y + 1] = id;
					self:GridArrange();
					return true;
				else
					table.remove(self.fellowship[x], y);
					x = x + orientDFellowship[self.orient];
					if self.fellowship[x] ~= nil then
						table.insert(self.fellowship[x], 1, id);
					elseif x == 0 then
						table.insert(self.fellowship, 1, { [1] = id });
					else
						self.fellowship[x] = { [1] = id };
					end
				end
				self:GridArrange();
				return true;
			end
		else
			if dy < 0 then
				table.remove(self.fellowship[x], y);
				if orientDFellowship[self.orient] > 0 then
					table.insert(self.fellowship, x, { [1] = id });
				else
					table.insert(self.fellowship, x + 1, { [1] = id });
				end
				self:GridArrange();
				return true;
			elseif dy > 0 then
				table.remove(self.fellowship[x], y);
				if orientDFellowship[self.orient] < 0 then
					table.insert(self.fellowship, x, { [1] = id });
				else
					table.insert(self.fellowship, x + 1, { [1] = id });
				end
				self:GridArrange();
				return true;
			end
		end
	else
		local major = sign(dx) * sign(self.majorX) + sign(dy) * sign(self.majorY);
		local minor = sign(dx) * sign(self.minorX) + sign(dy) * sign(self.minorY);

		if major ~= 0 then
			if self.grid[x][y].offset + major >= 0 then
				if self.grid[x][y + major] ~= nil and self.grid[x][y + major].offset == (self.grid[x][y].offset + major) then
					self.grid[x][y].name = self.grid[x][y + major].name;
					self.grid[x][y + major].name = id;
				else
					self.grid[x][y].offset = self.grid[x][y].offset + major;
				end
				self:GridArrange();
				return true;
			end
		elseif minor ~= 0 then
			if self.grid[x + minor] ~= nil then
				-- TODO
				for _, v in ipairs(self.grid[x + minor]) do
					if v.offset == self.grid[x][y].offset then
						self.grid[x][y].name = v.name;
						v.name = id;
						self:GridArrange();
						return true;
					end
				end

				table.insert(self.grid[x + minor], { offset = self.grid[x][y].offset, name = id });
				table.sort(self.grid[x + minor], function(a, b) return a.offset < b.offset; end);
				table.remove(self.grid[x], y);
				self:GridArrange();
				return true;
			else
				self.grid[x + minor] = { [1] = table.remove(self.grid[x], y) };
				self:GridArrange();
				return true;
			end
		end
	end
	return false;
end

function PartyFrameWindow:UpdateParty()
	for event, handle in pairs(self.partyHandle) do
		moebius92.Handlers.RemoveEventHandler(self.party, event, handle);
	end
	self.partyHandle = { };

	for _, actor in pairs(self.actors) do
		actor:RemoveHandlers();
		actor:SetVisible(false);
		actor:SetParent(nil);
	end
	self.players = { };
	self.actors = { };

	self.party = self.player:GetParty();
	self:PartyMemberChanged();
	if self.party ~= nil then
		self.partyHandle["MemberAdded"] = moebius92.Handlers.AddEventHandler(self.party, "MemberAdded", function(sender, args) self:PartyMemberChanged(); end);
		self.partyHandle["MemberRemoved"] = moebius92.Handlers.AddEventHandler(self.party, "MemberRemoved", function(sender, args) self:PartyMemberChanged(); end);
		self.partyHandle["LeaderChanged"] = moebius92.Handlers.AddEventHandler(self.party, "LeaderChanged", function(sender, args) self:PartyMemberChanged(); end);
	end
end

function PartyFrameWindow:PartyMemberChanged()
	self.actors = { };
	for _, player in pairs(self.players) do
		player.exists = false;
		player:SetStatus("Leader", false);
	end

	if self.party ~= nil and self.party:GetMemberCount() >= 1 then
		local n = self.party:GetMemberCount();
		local l = self.party:GetLeader();
		local name;

		for i = 1, n, 1 do
			local member = self.party:GetMember(i);
			name = member:GetName();

			self:AddMember(member, name);
			self.players[name].index = i;
			self.players[name]:SetStatus("Leader", l == member);
		end

		name = self.player:GetName();
		if self.players[name] == nil then
			self:AddMember(self.player, name);
			self.players[name].index = 0;
		end
	elseif fakeParty then
		local name = self.player:GetName();
		self:AddMember(self.player, name);
		self.players[name].index = 0;
	end

	self.nextUpdate = Turbine.Engine.GetGameTime();
	self:SetWantsUpdates(#self.actors > 0 and self.effectPoll > 0);

	for name, player in pairs(self.players) do
		if not player.exists then
			player:RemoveHandlers();
			player:SetVisible(false);
			player:SetParent(nil);
			if player.pet ~= nil then
				player.pet:RemoveHandlers();
				player.pet:SetVisible(false);
				player.pet:SetParent(nil);
			end
			self.players[name] = nil;
		else
			player.free = true;
		end
	end

	self:UpdateMorale();
	self:UpdatePower();
	for _, actor in pairs(self.actors) do
		actor:UpdateMorale();
		actor:UpdatePower();
	end

	if self.arrange == 1 then
		for i, group in ipairs(self.fellowship) do
			local newGroup = { };
			for _, name in ipairs(group) do
				if self.players[name] ~= nil and self.players[name].free then
					self.players[name].free = nil;
					table.insert(newGroup, name);
				end
			end

			self.fellowship[i] = newGroup;
		end

		for i = #self.fellowship, 1, -1 do
			if #self.fellowship[i] == 0 then
				table.remove(self.fellowship, i);
			end
		end

		local freePlayers = { };
		for name, player in pairs(self.players) do
			if player.free then
				player.free = nil;
				table.insert(freePlayers, { name = name, index = player.index });
			end
		end
		table.sort(freePlayers, function(a, b) return a.index < b.index; end);

		local n = #self.fellowship;
		if n == 0 then
			self.fellowship[1] = { };
			n = 1;
		end

		for _, player in ipairs(freePlayers) do
			table.insert(self.fellowship[n], player.name);
		end
	else
		-- TODO
		local t = { };
		local n;

		self.fellowship = { };

		for k, v in pairs(self.grid) do
			n = 0;
			t = { };
			for _, v1 in ipairs(v) do
				if self.players[v1.name] ~= nil and self.players[v1.name].free then
					self.players[v1.name].free = nil;
					table.insert(t, { offset = v1.offset - n, name = v1.name });
				else
					n = n + 1;
				end
			end
			self.grid[k] = t;
		end

		t = { };
		for k, v in pairs(self.players) do
			if v.free then
				v.free = nil;
				table.insert(t, { name = k, index = v.index });
			end
		end
		table.sort(t, function(a, b) return a.index < b.index; end);

		n = 0;
		for _, v in ipairs(t) do
			local offset;

			while self.grid[n] ~= nil and #self.grid[n] > 0 and self.grid[n][#self.grid[n]].offset >= self.maxOffset do
				if self.orient <= 2 then
					n = n + ((self.y < (self.h / 2)) and -1 or 1);
				else
					n = n + ((self.x < (self.w / 2)) and -1 or 1);
				end
			end

			if self.grid[n] == nil then
				self.grid[n] = { };
			end

			if #self.grid[n] == 0 then
				offset = 0;
			else
				offset = self.grid[n][#self.grid[n]].offset + 1;
			end

			table.insert(self.grid[n], { offset = offset, name = v.name});
		end
	end

	self:GridArrange();
end

function PartyFrameWindow:AddMember(member, name)
	if self.players[name] == nil then
		self.players[name] = PartyMemberWindow(self, member, self.layout, nil);
		self.players[name]:SetStatus("Leader", false);
		if self.arrange == 1 then
			self.players[name].showPet = 2;
		else
			if self.orient <= 2 then
				if self.x < (self.w / 2) then
					self.players[name].showPet = 5;
				else
					self.players[name].showPet = 4;
				end
			else
				if self.y < (self.h / 2) then
					self.players[name].showPet = 2;
				else
					self.players[name].showPet = 3;
				end
			end
		end
	end

	self.players[name].id = name;

	if not self.players[name].exists then
		self.players[name].exists = true;
		self.players[name]:SetVisible(true);
		self.players[name].select:SetMouseVisible(not self.move);
		table.insert(self.actors, self.players[name]);

		local pet = member:GetPet();

		if self.players[name].showPet == 1 then
			pet = nil;
		end

		if self.players[name].pet == nil or self.players[name].pet.actor ~= pet then
			if self.players[name].pet ~= nil then
				self.players[name].pet:RemoveHandlers();
				self.players[name].pet:SetVisible(false);
				self.players[name].pet:SetParent(nil);
				self.players[name].pet = nil;
			end

			if pet ~= nil then
				self.players[name].pet = PartyMemberWindow(self, pet, self.layout, self.players[name].class);
			end
		end

		if self.players[name].pet ~= nil then
			self.players[name].pet:SetVisible(true);
			table.insert(self.actors, self.players[name].pet);
		end
	end
end

function PartyFrameWindow:GridArrange()
	if self.arrange == 1 then
		for i = #self.fellowship, 1, -1 do
			if #self.fellowship[i] == 0 then
				table.remove(self.fellowship, i);
			end
		end

		self.grid = { };
		self.grid[0] = { };
		n = 0;
		for k, v in ipairs(self.fellowship) do
			local l = 0;

			if self.fellowtabs[k] == nil then
				self.fellowtabs[k] = PartyFrameFellowshipTab(self, k);
			end
			self.fellowtabs[k].s = n;

			if self.orient == 1 then
				for i = #v, 1, -1 do
					l = l + 1;
					if self.players[v[i]].pet then
						n = n + 1;
						l = l + 1;
					end
					table.insert(self.grid[0], { offset = n, name = v[i] });
					self.players[v[i]].indexX = k;
					self.players[v[i]].indexY = i;
					n = n + 1;
				end
			else
				for i = 1, #v, 1 do
					l = l + 1;
					table.insert(self.grid[0], { offset = n, name = v[i] });
					self.players[v[i]].indexX = k;
					self.players[v[i]].indexY = i;
					if self.players[v[i]].pet then
						n = n + 1;
						l = l + 1;
					end
					n = n + 1;
				end
			end

			self.fellowtabs[k].l = l;

			n = n + 1;
		end
	else
		for k, v in pairs(self.grid) do
			for k1, v1 in ipairs(v) do
				self.players[v1.name].indexX = k;
				self.players[v1.name].indexY = k1;
			end
		end
	end

	for k, v in pairs(self.grid) do
		if #v == 0 then
			self.grid[k] = nil;
		end
	end

	self:Arrange();
end

function PartyFrameWindow:Arrange()
	local h, w = self:GetSize();
	local petDX = { nil, 0, 0, -1, 1 };
	local petDY = { nil, 1, -1, 0, 0 };

	self:SetPosition(self.x + self.dX, self.y + self.dY);

	if self.arrange == 1 then
		local o = self.x < (self.w / 2);

		for i = 1, #self.fellowship, 1 do
			self.fellowtabs[i]:Orient(o);
			if self.orient == 1 then
				self.fellowtabs[i]:SetPosition(self.x + (o and self.minorX or -self.size), self.y + (self.fellowtabs[i].s + self.fellowtabs[i].l - 1) * self.majorY);
			else
				self.fellowtabs[i]:SetPosition(self.x + (o and self.minorX or -self.size), self.y + self.fellowtabs[i].s * self.majorY);
			end
		end
		for i = #self.fellowship + 1, #self.fellowtabs, 1 do
			self.fellowtabs[i]:SetVisible(false);
		end

		if not self.move and self.grid[0] ~= nil and #self.grid[0] > 0 then
			local n = self.grid[0][#self.grid[0]].offset;

			if self.orient == 1 then
				if n >= self.maxOffset then
					for _, v in pairs(self.grid[0]) do
						self.players[v.name]:SetPosition(self.x, 0 + (v.offset - n) * self.majorY);
						if self.players[v.name].pet ~= nil then
							self.players[v.name].pet:SetPosition(self.x, 0 + (v.offset - 1 - n) * self.majorY);
						end
					end
					return;
				end
			else
				if self.grid[0][#self.grid[0]].pet ~= nil then
					n = n + 1;
				end

				if n >= self.maxOffset then
					for _, v in pairs(self.grid[0]) do
						self.players[v.name]:SetPosition(self.x, self.h + (v.offset - n) * self.majorY);
						if self.players[v.name].pet ~= nil then
							self.players[v.name].pet:SetPosition(self.x, self.h + (v.offset + 1 - n) * self.majorY);
						end
					end
					return;
				end
			end
		end
	else
		for i = 1, #self.fellowtabs, 1 do
			self.fellowtabs[i]:SetVisible(false);
		end
	end

	for k, v in pairs(self.grid) do
		for _, v1 in pairs(v) do
			self.players[v1.name]:SetPosition(self.x + k * self.minorX + v1.offset * self.majorX,
											  self.y + k * self.minorY + v1.offset * self.majorY);
			if self.players[v1.name].pet ~= nil then
				self.players[v1.name].pet:SetPosition(self.x + k * self.minorX + v1.offset * self.majorX + petDX[self.players[v1.name].showPet] * math.abs(self.majorX + self.minorX),
													  self.y + k * self.minorY + v1.offset * self.majorY + petDY[self.players[v1.name].showPet] * math.abs(self.majorY + self.minorY));
			end
		end
	end
end

function PartyFrameWindow:ChatMessage(m)
	local x, _, name, roll = string.find(m, S.RollMessage);

	if x == 1 and self.players[name] ~= nil and tonumber(roll) ~= nil then
		local t = { };
		local i, j, k;
		local state, w;

		table.insert(self.players[name].rolls, tonumber(roll));

		for name, player in pairs(self.players) do
			table.insert(t, { rolls = player.rolls, name = name });
		end
		table.sort(t,
function(a, b)
	for i = 1, math.min(#a.rolls, #b.rolls), 1 do
		if a.rolls[i] > b.rolls[i] then
			return true;
		elseif a.rolls[i] < b.rolls[i] then
			return false;
		end
	end

	return #a.rolls > #b.rolls;
end
		);

		k = 0;
		for i = 1, #t, 1 do
			if #t[i].rolls == 0 then
				break;
			end

			state = 0;
			if i == 1 then
				state = 2
			else
				for j = 1, #t[i].rolls, 1 do
					if t[i - 1].rolls[j] > t[i].rolls[j] then
						state = 2;
						break;
					end
				end

				if state == 0 then
					if #t[i - 1].rolls > #t[i].rolls then
						state = 1;
					end
				end
			end

			if state == 2 then
				k = k + 1;
				w = k;
			elseif state == 1 then
				w = k .. "?";
			else
				w = k;
			end

			if self.players[t[i].name].rollOrder ~= w then
				self.players[t[i].name].rollOrder = w;
				self.players[t[i].name]:UpdateName();
			end
		end
	elseif string.find(m, S.JoinMessage) == 1 then
		if not self.reload then
			self.reload = true;

			Turbine.PluginManager.LoadPlugin("PartyFrameReloader");
		end
	end
end

function PartyFrameWindow:ClearRolls()
	for _, player in pairs(self.players) do
		player.rolls = { };
		player.rollOrder = nil;
		player:UpdateName();
	end
end

-- TODO
function PartyFrameWindow:EffectControl(debuff, icon, name, category, curable)
	local t = { };

	debuff = debuff and "D" or "B";

	r = self.effectControl[debuff .. "/" .. icon .. "/" .. name];
	if r ~= nil then
		for _, v in pairs(r) do
			if t[v.control] == nil or t[v.control] > v.group then
				t[v.control] = v.group;
			end
		end
	end

	r = self.effectControl[debuff .. "/" .. name];
	if r ~= nil then
		for _, v in pairs(r) do
			if t[v.control] == nil or t[v.control] > v.group then
				t[v.control] = v.group;
			end
		end
	end

	r = self.effectControl[debuff .. "/" .. icon];
	if r ~= nil then
		for _, v in pairs(r) do
			if t[v.control] == nil or t[v.control] > v.group then
				t[v.control] = v.group;
			end
		end
	end

	if debuff and category ~= 0 then
		curable = curable and "Cure" or "ICure";
		r = self.effectControl[debuff .. "/" .. S["SE_" .. curable]];
		if r ~= nil then
			for _, v in pairs(r) do
				if t[v.control] == nil or t[v.control] > v.group then
					t[v.control] = v.group;
				end
			end
		end

		for k, v in pairs(Turbine.Gameplay.EffectCategory) do
			if S["SE_" .. k] ~= nil and (v % (2 * category)) >= category then
				r = self.effectControl[debuff .. "/" .. S["SE_" .. k]];
				if r ~= nil then
					for _, v in pairs(r) do
						if t[v.control] == nil or t[v.control] > v.group then
							t[v.control] = v.group;
						end
					end
				end

				r = self.effectControl[debuff .. "/" .. S["SE_" .. k .. curable]];
				if r ~= nil then
					for _, v in pairs(r) do
						if t[v.control] == nil or t[v.control] > v.group then
							t[v.control] = v.group;
						end
					end
				end
			end
		end
	end

	return t;
end

function PartyFrameWindow:MoraleControl(abs, per)
	local t = { };

	for k, v in ipairs(self.moraleControl) do
		if v.f(abs, per) then
			if t[v.control] == nil or t[v.control] > v.group then
				t[v.control] = v.group;
			end
		end
	end

	return t;
end

function PartyFrameWindow:PowerControl(abs, per, status)
	local t = { };

	for k, v in ipairs(self.powerControl) do
		if v.f(abs, per) then
			if t[v.control] == nil or t[v.control] > v.group then
				t[v.control] = v.group;
			end
		end
	end
	return t;
end

function PartyFrameWindow:StatusControl(status, v)
	local t = { };

	if v ~= nil then
		status = status .. "_" .. tostring(v);
	end

	status = "SE_" .. status;
	if S[status] ~= nil then
		r = self.effectControl[S[status]];
		if r ~= nil then
			for _, v in pairs(r) do
				if t[v.control] == nil or t[v.control] > v.group then
					t[v.control] = v.group;
				end
			end
		end
	end
	return t;
end

function PartyFrameWindow:LogEffect(effect)
	local debuff = effect:IsDebuff();
	local icon = effect:GetIcon();
	local name = effect:GetName();

	if debuff then
		if self.options.logDebuffs.index == 1 then
			return;
		elseif self.options.logDebuffs.index < 4 then
			if self.effectControl["D/" .. icon .. "/" .. name] == nil and self.effectControl["D/" .. name] ~= nil then
			elseif next(self:EffectControl(debuff, icon, name, effect:GetCategory(), effect:IsCurable())) ~= nil then
				return;
			elseif self.options.logDebuffs.index == 2 then
				return;
			end
		end
	else
		if self.options.logBuffs.index == 1 then
			return;
		elseif self.options.logBuffs.index < 4 then
			if self.effectControl["B/" .. icon .. "/" .. name] == nil and self.effectControl["B/" .. name] ~= nil then
			elseif next(self:EffectControl(debuff, icon, name, effect:GetCategory(), effect:IsCurable())) ~= nil then
				return;
			elseif self.options.logBuffs.index == 2 then
				return;
			end
		end
	end

	for _, v in ipairs(self.options.controls[0]) do
		if v.name == name and v.debuff == debuff and v.icon == icon then
			return;
		end
	end

	table.insert(self.options.controls[0], { name = name, debuff = debuff, icon = icon });
	table.sort(self.options.controls[0], function(a, b) return a.name < b.name; end);
	self.options:Redraw();
end

function PartyFrameWindow:PetChanged()
	if self.pet ~= nil then
		moebius92.Handlers.RemoveEventHandler(self.petEffects, "EffectAdded", self.petHandler);
		self.pet = nil;
	end

	self.pet = self.player:GetPet();

	if self.pet ~= nil then
		self.petEffects = self.pet:GetEffects();
		self.petHandler = moebius92.Handlers.AddEventHandler(self.petEffects, "EffectAdded", function(sender, args) self:LogEffect(self.petEffects:Get(args.Index)); end);
	end
end

function PartyFrameWindow:CombatChange()
	self.inCombat = self.player:IsInCombat();
	for _, actor in pairs(self.actors) do
		actor:UpdateMorale();
		actor:UpdatePower();
	end
end

function PartyFrameWindow:UpdateMorale()
	local m = 1;

	for _, actor in pairs(self.actors) do
		m = math.max(m, actor.maxMorale);
	end

	if m ~= self.maxMorale then
		self.maxMorale = m;
		for _, actor in pairs(self.actors) do
			actor:UpdateMorale();
		end
	end
end

function PartyFrameWindow:UpdatePower()
	local m = 1;

	for _, actor in pairs(self.actors) do
		m = math.max(m, actor.maxPower);
	end

	if m ~= self.maxPower then
		self.maxPower = m;
		for _, actor in pairs(self.actors) do
			actor:UpdatePower();
		end
	end
end

function PartyFrameWindow:UpdateControls()
	local l;

	if self.party ~= nil then
		l = self.party:GetLeader();
	end

	for _, actor in pairs(self.actors) do
		actor:InitializeControls();
		actor:SetStatus("Leader", l == actor.actor);
	end
end

function PartyFrameWindow:ClearMembers()
	for _, actor in pairs(self.actors) do
		actor:RemoveHandlers();
		actor:SetVisible(false);
		actor:SetParent(nil);
	end
	self.players = { };
	self.actors = { };
	self.fellowship = { };
	self.grid = { };
end

function PartyFrameWindow:SetLayout(layout)
	self.layout = layout;
	Turbine.UI.Lotro.LotroUI.SetEnabled(Turbine.UI.Lotro.LotroUIElement.Party, layout.overlay);
	self:PartyMemberChanged();
	self:Orient();
	self:SaveSettings();
end