PartyMemberWindow = class(Turbine.UI.Window);

local tempMoraleIcon = 0x411188D9;

function PartyMemberWindow:Constructor(parent, actor, layout, petclass)
	Turbine.UI.Window.Constructor(self);

	self.parent = parent;
	self.actor = actor;
	self.petclass = petclass;
	if self.petclass == nil then
		self.class = actor:GetClass();
	else
		self.class = petclass;
	end

	self.rolls = { };
	self.rollOrder = nil;
	self.handles = { };

	if layout.pre ~= nil then
		layout.pre(self);
	end

	self:SetSize(layout.width, layout.height);
	self.controls = { };
	for i, control in ipairs(layout.layout) do
		if control.element ~= nil then
			self.controls[i] = control.element(self, control.settings);
		else
			self.controls[i] = { };
		end
		self.controls[i].index = i;
	end

	if layout.post ~= nil then
		layout.post(self);
	end

	self.select = Turbine.UI.Lotro.EntityControl();
	self.select:SetParent(self);
	self.select:SetPosition(0, 0);
	self.select:SetMouseVisible(true);
	self.select:SetEntity(actor);
	self.select:SetSelectionEnabled(true);
	self.select:SetContextMenuEnabled(true);
	self.select:SetSize(self:GetSize());
	self:SetMouseVisible(true);

	self.handles[self.actor] = { };
	self.handles[self.actor].NameChanged = moebius92.Handlers.AddEventHandler(self.actor, "NameChanged", function(sender, args) self:UpdateName(); end);

	self.oldMorale = actor:GetMorale();
	self.tempMoraleEffect = { };
	self.tempMoraleEnd = 0;
	self.handles[self.actor].MoraleChanged = moebius92.Handlers.AddEventHandler(self.actor, "MoraleChanged", function(sender, args) self:UpdateMorale(); end);
	self.handles[self.actor].MaxMoraleChanged = moebius92.Handlers.AddEventHandler(self.actor, "MaxMoraleChanged", function(sender, args) self:UpdateMorale(); end);
	self.handles[self.actor].BaseMaxMoraleChanged = moebius92.Handlers.AddEventHandler(self.actor, "BaseMaxMoraleChanged", function(sender, args) self:UpdateMorale(); end);
	self.handles[self.actor].TemporaryMoraleChanged = moebius92.Handlers.AddEventHandler(self.actor, "TemporaryMoraleChanged", function(sender, args) self:UpdateMorale(); end);
	self.handles[self.actor].MaxTemporaryMoraleChanged = moebius92.Handlers.AddEventHandler(self.actor, "MaxTemporaryMoraleChanged", function(sender, args) self:UpdateMorale(); end);

	self.oldPower = actor:GetPower();
	self.tempPowerEffect = { };
	self.tempPowerEnd = 0;
	self.handles[self.actor].PowerChanged = moebius92.Handlers.AddEventHandler(self.actor, "PowerChanged", function(sender, args) self:UpdatePower(); end);
	self.handles[self.actor].MaxPowerChanged = moebius92.Handlers.AddEventHandler(self.actor, "MaxPowerChanged", function(sender, args) self:UpdatePower(); end);
	self.handles[self.actor].BaseMaxPowerChanged = moebius92.Handlers.AddEventHandler(self.actor, "BaseMaxPowerChanged", function(sender, args) self:UpdatePower(); end);
	self.handles[self.actor].TemporaryPowerChanged = moebius92.Handlers.AddEventHandler(self.actor, "TemporaryPowerChanged", function(sender, args) self:UpdatePower(); end);
	self.handles[self.actor].MaxTemporaryPowerChanged = moebius92.Handlers.AddEventHandler(self.actor, "MaxTemporaryPowerChanged", function(sender, args) self:UpdatePower(); end);

	self.effects = self.actor:GetEffects();
	self.handles[self.effects] = { };
	self.handles[self.effects].EffectAdded = moebius92.Handlers.AddEventHandler(self.effects, "EffectAdded", function(sender, args) self:AddEffect(self.effects:Get(args.Index)); end);
	self.handles[self.effects].EffectRemoved = moebius92.Handlers.AddEventHandler(self.effects, "EffectRemoved", function(sender, args) self:RemoveEffect(args.Effect, true); end);
	self.handles[self.effects].EffectsCleared = moebius92.Handlers.AddEventHandler(self.effects, "EffectsCleared", function(sender, args) self:ClearEffects(); end);

	if self.petclass == nil then
		self.handles[self.actor].IsLinkDeadChanged = moebius92.Handlers.AddEventHandler(self.actor, "IsLinkDeadChanged", function(sender, args) self:SetStatus("LinkDead", self.actor:IsLinkDead()); end);
		self.handles[self.actor].IsVoiceEnabledChanged = moebius92.Handlers.AddEventHandler(self.actor, "IsVoiceEnabledChanged", function(sender, args) self:SetStatus("Voice", self.actor:IsVoiceEnabled()); end);
		self.handles[self.actor].IsVoiceActiveChanged = moebius92.Handlers.AddEventHandler(self.actor, "IsVoiceActiveChanged", function(sender, args) self:SetStatus("Talking", self.actor:IsVoiceActive()); end);
		self.handles[self.actor].ReadyStateChanged = moebius92.Handlers.AddEventHandler(self.actor, "ReadyStateChanged", function(sender, args) self:SetStatus("Ready", self.actor:GetReadyState()); end);
		self.handles[self.actor].PetChanged = moebius92.Handlers.AddEventHandler(self.actor, "PetChanged", function(sender, args) self.parent:PartyMemberChanged(); end);
	end

	self:InitializeControls();

	self:SetVisible(self.parent.hide);
end

function PartyMemberWindow:RemoveHandlers()
	for object, events in pairs(self.handles) do
		for event, handle in pairs(events) do
			moebius92.Handlers.RemoveEventHandler(object, event, handle);
		end
	end
end

function PartyMemberWindow:MouseDown(args)
	if args.Button == Turbine.UI.MouseButton.Left then
		self.dragging = true;
		self.safeY = nil;
	end
end

function PartyMemberWindow:MouseMove(args)
	local w, h = self:GetSize();

	if self.dragging then
		local x, y;

		if not IsSafeY(self, args.Y) then
			return;
		end

		x = math.floor(args.X / w);
		y = math.floor(args.Y / h);
		if x == 0 and y == 1 and self.parent.arrange == 1 and self.pet ~= nil then
			y = 0;
		end

		if x ~= 0 or y ~= 0 then
			if self.parent:MoveMember(self.id, x, y) and self.parent.arrange == 1 then
				if y < 0 then
					self.safeY = math.huge;
				else
					self.safeY = -math.huge;
				end
			end
		end
	end
end

function PartyMemberWindow:MouseUp(args)
	if args.Button == Turbine.UI.MouseButton.Left then
		self.dragging = false;
	end
end

function PartyMemberWindow:MouseClick(args)
	if args.Button == Turbine.UI.MouseButton.Right then
		self.parent.memberSelect = self;
		for i = 1, #self.parent.memberOptions, 1 do
			self.parent.memberOptions[i]:SetChecked(i == self.showPet);
			if self.parent.arrange == 1 then
				self.parent.memberOptions[i]:SetEnabled(i <= 2);
			else
				self.parent.memberOptions[i]:SetEnabled(true);
			end
		end
		self.parent.memberMenu:ShowMenu();
	end
end

function PartyMemberWindow:UpdateName()
	local name = self.actor:GetName();

	if name == nil or name == "" then
		self.pollName = true;
		self:SetWantsUpdates(true);
	else
		self.pollName = false;
	end

	for _, control in pairs(self.controls) do
		if control.SetName ~= nil then
			control:SetName(name, self.rollOrder);
		end
	end
	self:CheckTarget();
end

function PartyMemberWindow:CheckTarget()
	self:SetStatus("Target", self.parent.target ~= nil and self.parent.target:GetName() == self.actor:GetName());
end

function PartyMemberWindow:UpdateMorale()
	local cur = self.actor:GetMorale();
	local base = self.actor:GetBaseMaxMorale();
	local max = self.actor:GetMaxMorale();
	local temp = self.actor:GetTemporaryMorale();
	local tmax = self.actor:GetMaxTemporaryMorale();

	local newCtrlGrp = self.parent:MoraleControl(cur, cur / max);
	local deltaCtrlGrp = { };
	for control, group in pairs(self.moraleState) do
		if newCtrlGrp[control] ~= group then
			deltaCtrlGrp[control] = group;
		end
	end
	self:DecrementControlGroup(deltaCtrlGrp);
	deltaCtrlGrp = { };
	for control, group in pairs(newCtrlGrp) do
		if self.moraleState[control] ~= group then
			deltaCtrlGrp[control] = group;
		end
	end
	self:IncrementControlGroup(deltaCtrlGrp);
	self.moraleState = newCtrlGrp;

	for _, control in pairs(self.controls) do
		if control.SetMorale ~= nil then
			control:SetMorale(cur, base, max, temp, tmax, self.parent.maxMorale, self.oldMorale, self.parent.inCombat, self.tempMoraleEnd);
		end
	end

	self.maxMorale = math.max(max, base) + math.min(tmax, base);
	self.oldMorale = cur;
	self.parent:UpdateMorale();
end

function PartyMemberWindow:UpdatePower()
	local cur = self.actor:GetPower();
	local base = self.actor:GetBaseMaxPower();
	local max = self.actor:GetMaxPower();
	local temp = self.actor:GetTemporaryPower();
	local tmax = self.actor:GetMaxTemporaryPower();

	-- Treat Beornings as if they have one power
	if self.class == Turbine.Gameplay.Class.Beorning then
		cur = 1;
		base = 1;
		max = 1;
		temp = 0;
		tmax = 0;
	end

	local newCtrlGrp = self.parent:PowerControl(cur, cur / max);
	local deltaCtrlGrp = { };
	for control, group in pairs(self.powerState) do
		if newCtrlGrp[control] ~= group then
			deltaCtrlGrp[control] = group;
		end
	end
	self:DecrementControlGroup(deltaCtrlGrp);
	deltaCtrlGrp = { };
	for control, group in pairs(newCtrlGrp) do
		if self.powerState[control] ~= group then
			deltaCtrlGrp[control] = group;
		end
	end
	self:IncrementControlGroup(deltaCtrlGrp);
	self.powerState = newCtrlGrp;

	for _, control in pairs(self.controls) do
		if control.SetPower ~= nil then
			control:SetPower(cur, base, max, temp, tmax, self.parent.maxPower, self.oldPower, self.parent.inCombat, self.tempPowerEnd);
		end
	end

	self.maxPower = math.max(max, base) + math.min(tmax, base);
	self.oldPower = cur;
	self.parent:UpdatePower();
end

function PartyMemberWindow:SetStatus(status, v)
	if v == nil or self.statusState[status] ~= v then
		local t;

		if self.statusState[status] ~= nil then
			t = self.parent:StatusControl(status, self.statusState[status]);
			self:DecrementControlGroup(t);
		end

		t = self.parent:StatusControl(status, v);
		self:IncrementControlGroup(t);

		self.statusState[status] = v;
	end
end

function PartyMemberWindow:AddEffect(effect)
	local id;
	local time = Turbine.Engine.GetGameTime();
	local name = effect:GetName();
	local icon = effect:GetIcon();

	self.parent:LogEffect(effect);

	self.lastEffectUpdate = time;
	for i = 1, self.effects:GetCount(), 1 do
		id = self.effects:Get(i):GetID();

		if self.effectState[id] ~= nil then
			self.effectState[id].time = time;
		end
	end
	self:SetWantsUpdates(true);

	id = effect:GetID();

	if icon == tempMoraleIcon or tempMoraleEffects[name] then
		self.tempMoraleEffect[id] = effect:GetStartTime() + effect:GetDuration();
		self.tempMoraleEnd = 0;
		for _, endTime in pairs(self.tempMoraleEffect) do
			self.tempMoraleEnd = math.max(self.tempMoraleEnd, endTime);
		end
		self:UpdateMorale();
	end

	if self.effectState[id] == nil then
		local ctrlGrp = self.parent:EffectControl(effect:IsDebuff(), icon, name, effect:GetCategory(), effect:IsCurable());

		if next(ctrlGrp) ~= nil then
			local item = { };

			item.effect = effect;
			item.time = time;
			item.controls = ctrlGrp;
			item.startTime = effect:GetStartTime();
			item.durationTime = effect:GetDuration();
			item.endTime = item.startTime + item.durationTime;
			self.effectState[id] = item;

			self:IncrementControlGroup(ctrlGrp);
			for control, group in pairs(ctrlGrp) do
				if self.controls[control].effects[group] == nil then
					self.controls[control].effects[group] = { };
				end
				table.insert(self.controls[control].effects[group], item );
				table.sort(self.controls[control].effects[group], function(a, b) return a.startTime > b.startTime; end);
				self:UpdateControlEffect(control);
			end
		end
	end
end

function PartyMemberWindow:RemoveEffect(effect, mark)
	local id;
	local time = Turbine.Engine.GetGameTime();
	local name = effect:GetName();

	if mark then
		self.lastEffectUpdate = time;
		for i = 1, self.effects:GetCount(), 1 do
			id = self.effects:Get(i):GetID();

			if self.effectState[id] ~= nil then
				self.effectState[id].time = time;
			end
		end
	end

	id = effect:GetID();

	if self.tempMoraleEffect[id] ~= nil then
		self.tempMoraleEffect[id] = nil;
		self.tempMoraleEnd = 0;
		for _, endTime in pairs(self.tempMoraleEffect) do
			self.tempMoraleEnd = math.max(self.tempMoraleEnd, endTime);
		end
	end

	if self.effectState[id] ~= nil then
		self:DecrementControlGroup(self.effectState[id].controls);
		for control in pairs(self.effectState[id].controls) do
			for group, effects in pairs(self.controls[control].effects) do
				for i, item in ipairs(effects) do
					if item == self.effectState[id] then
						table.remove(self.controls[control].effects[group], i);
						if #self.controls[control].effects[group] == 0 then
							self.controls[control].effects[group] = nil;
						end
						self:UpdateControlEffect(control);
						break;
					end
				end
			end
		end
		self.effectState[id] = nil;
	end
end

function PartyMemberWindow:ClearEffects()
	for _, effect in pairs(self.effectState) do
		self:DecrementControlGroup(effect.controls);
	end
	self.effectState = { };

	for control in ipairs(self.controls) do
		self.controls[control].effects = { };
		self:UpdateControlEffect(control);
	end

	self.tempMoraleEffect = { };
	self.tempMoraleEnd = 0;
	self:UpdateMorale();

	self.tempPowerEffect = { };
	self.tempPowerEnd = 0;
	self:UpdatePower();
end

function PartyMemberWindow:CheckEffects(effectIDs)
	for id, effect in pairs(self.effectState) do
		if not effectIDs[id] then
			self:RemoveEffect(effect.effect, false);
		end
	end
end

function PartyMemberWindow:Update()
	local time = Turbine.Engine.GetGameTime();
	local update = false;

	if self.lastEffectUpdate == time then
		for _, effect in pairs(self.effectState) do
			if effect.time ~= time then
				self:RemoveEffect(effect.effect, false);
			end
		end
	else
		for _, effect in pairs(self.effectState) do
			if effect.endTime <= time then
				self:RemoveEffect(effect.effect, false);
			end
		end
	end

	for _, effect in pairs(self.effectState) do
		if effect.durationTime < 1000000 then
			update = true;
			break;
		end
	end

	if self.pollName then
		self:UpdateName();
		update = true;
	end

	self:SetWantsUpdates(update);
end

function PartyMemberWindow:InitializeControls()
	self.effectState = { };
	self.statusState = { };
	self.moraleState = { };
	self.powerState = { };
	for _, control in ipairs(self.controls) do
		control.effects = { };
		control.group = { };
	end

	self:UpdateName();
	self:UpdateMorale();
	self:UpdatePower();

	self.lastEffectUpdate = 0;
	for i = 1, self.effects:GetCount(), 1 do
		self:AddEffect(self.effects:Get(i));
	end

	if self.petclass == nil then
		self:SetStatus("LinkDead", self.actor:IsLinkDead());
		self:SetStatus("Voice", self.actor:IsVoiceEnabled());
		self:SetStatus("Talking", self.actor:IsVoiceActive());
		self:SetStatus("Ready", self.actor:GetReadyState());
	end
	self:SetStatus("Default", nil);
	if self.petclass == nil then
		self:SetStatus("Player", nil);
	else
		self:SetStatus("Pet", nil);
	end
	for name, x in pairs(Turbine.Gameplay.Class) do
		if x == self.class then
			self:SetStatus(name, nil);
			break;
		end
	end

	for control in ipairs(self.controls) do
		self:UpdateControlColor(control);
		self:UpdateControlEffect(control);
	end
	self:CheckTarget();
end

function PartyMemberWindow:IncrementControlGroup(ctrlGrp)
	for control, group in pairs(ctrlGrp) do
		self.controls[control].group[group] = (self.controls[control].group[group] or 0) + 1;

		local controlOption = self.parent.options.controls[control];
		if controlOption.color ~= nil and controlOption.color[group] ~= nil then
			if self.controls[control].group[group] == math.max(1, controlOption.color[group].count or 1) then
				self:UpdateControlColor(control);
			end
		end
	end
end

function PartyMemberWindow:DecrementControlGroup(ctrlGrp)
	for control, group in pairs(ctrlGrp) do
		self.controls[control].group[group] = self.controls[control].group[group] - 1;

		local controlOption = self.parent.options.controls[control];
		if controlOption.color ~= nil and controlOption.color[group] ~= nil then
			if self.controls[control].group[group] == (math.max(1, controlOption.color[group].count or 1) - 1) then
				self:UpdateControlColor(control);
			end
		end

		if self.controls[control].group[group] <= 0 then
			self.controls[control].group[group] = nil;
		end
	end
end

function PartyMemberWindow:UpdateControlColor(control)
	local controlOption = self.parent.options.controls[control];

	if self.controls[control].SetColor ~= nil and controlOption.color ~= nil then
		local H, S, V;

		for i, group in ipairs(controlOption.color) do
			if controlOption.missing == ((self.controls[control].group[i] or 0) < math.max(1, group.count or 1)) then
				if group.color.V ~= nil then
					if V == nil then
						V = group.color.V;
					else
						V = (V + group.color.V) / 2;
					end
				end

				if group.color.H ~= nil then
					if H == nil then
						H = group.color.H;
						S = group.color.S;
					else
						local r = group.color.S / 2;
						local theta = (math.acos(-r / 2) - 2 * math.pi / 3) * S + (2 * math.pi / 3);
						local H1 = (group.color.H - H) % 6;

						if H1 <= 2 then
							theta = theta * H1 / 2;
						elseif H1 <= 4 then
							theta = theta + theta * (H1 - 2);
						else
							theta = -theta * (6 - H1) / 2;
						end

						local x = math.cos(theta) * r + S;
						local y = math.sin(theta) * r;
						r = math.sqrt(x * x + y * y);
						if r == 0 then
							H = 0.0;
							S = 0.0;
						else
							H = ((math.atan(y / r, x / r) * 3 / math.pi) + H) % 6;
							S = math.min(r, 1.0);
						end
					end
				end

				if group.color.H ~= nil and group.color.V ~= nil then
					break;
				end
			end
		end

		if H == nil then
			self.controls[control]:SetColor(nil, nil, nil);
		else
			local c = { };

			c[0] = 1.0 - S;
			c[1] = 1.0 - S
			c[2] = 1.0 - S
			H = (H + 1) % 6;
			x = math.floor(H / 2);
			c[x] = c[x] + S;
			H = H % 2;
			if H < 1 then
				x = (x - 1) % 3;
				c[x] = c[x] + S * (1.0 - H);
			else
				x = (x + 1) % 3;
				c[x] = c[x] + S * (H - 1.0);
			end
			V = V and V or 1.0;
			self.controls[control]:SetColor(c[0] * V, c[1] * V, c[2] * V);
		end
	end
end

function PartyMemberWindow:UpdateControlEffect(control)
	local controlOption = self.parent.options.controls[control];

	if controlOption.missing then
		if self.controls[control].SetIcon ~= nil then
			if controlOption.color ~= nil then
				local icon = nil;

				for i, group in ipairs(controlOption.color) do
					if (self.controls[control].group[i] or 0) < math.max(1, group.count or 1) then
						if group[1] ~= nil then
							icon = group[1].icon;
						end
						break;
					end
				end

				if icon == 0 then
					icon = nil;
				end

				self.controls[control]:SetIcon(icon);
			else
				self.controls[control]:SetIcon(nil);
			end
		end
	else
		if self.controls[control].SetEffect ~= nil then
				if controlOption.color ~= nil then
				local groups = { };

				for group in pairs(self.controls[control].effects) do
					table.insert(groups, group);
				end
				table.sort(groups);

				for _, group in ipairs(groups) do
					local count = controlOption.color[group].count;
					local n = #self.controls[control].effects[group];

					if n > 0 then
						if count == nil or count == 0 then
							self.controls[control]:SetEffect(self.controls[control].effects[group][1].effect, n);
							return;
						elseif count < 0 and -count <= n then
							self.controls[control]:SetEffect(self.controls[control].effects[group][n + count + 1].effect);
							return;
						elseif count > 0 and count <= n then
							self.controls[control]:SetEffect(self.controls[control].effects[group][count].effect);
							return;
						end
					end
				end
				self.controls[control]:SetEffect(nil);
			else
				self.controls[control]:SetEffect(nil);
			end
		elseif self.controls[control].SetIcon ~= nil then
			if controlOption.color ~= nil then
				local groups = { };

				for group in pairs(self.controls[control].effects) do
					table.insert(groups, group);
				end
				table.sort(groups);

				for _, group in ipairs(groups) do
					local count = controlOption.color[group].count;
					local n = #self.controls[control].effects[group].effects;

					if n > 0 then
						if count == nil or count == 0 then
							self.controls[control]:SetIcon(self.controls[control].effects[group][1].effect:GetIcon(), n);
							return;
						elseif count < 0 and -count <= n then
							self.controls[control]:SetIcon(self.controls[control].effects[group][n + count + 1].effect:GetIcon());
							return;
						elseif count > 0 and count <= n then
							self.controls[control]:SetIcon(self.controls[control].effects[group][count].effect:GetIcon());
							return;
						end
					end
				end
				self.controls[control]:SetIcon(nil);
			else
				self.controls[control]:SetIcon(nil);
			end

		end
	end
end